/*************************************************************************************************************************************************
 * Nano servo driver
 *
 *************************************************************************************************************************************************/

#ifndef servoedit4_h
#define servoedit4_h

#include "Servo.h"

const int NumServosMax = 12;

uint8_t NumServos = 1;

struct TServo {
    uint8_t Group;
    uint8_t PinNum;
    uint16_t ServoMin,ServoMax;
    float Target;
    float Position;
    float Velocity;
    float Vmax;
    float Amax;
    Servo aServo;
  };

TServo Servos[NumServosMax];

bool MoveServos();
void SetServoPosition(uint8_t servo, int16_t pos);

bool AtTargets = true;

#define SERVO_FREQ 50 // 50 Hz updates

const int VelPeriod = 50; // how often the servo position is recalculated

//-----------------------------------------------------------
// sign()
//   sign of number
//   a function not a macro so you don't calulate the parameter twice
//-----------------------------------------------------------
float sign(float a) {
  if (a < 0) return -1;
  if (a > 0) return +1;
  return 0;
}

//-----------------------------------------------------------
// sqr()
//   square funtion
//   it's a function not a macro like sq() 
//   so args are not calculated twice
//-----------------------------------------------------------
float sqr(float a) {
  return a*a;
}

//-----------------------------------------------------------
// Min()
//   Min funtion
//   it's a function not a macro like min() 
//   so args are not calculated twice
//-----------------------------------------------------------
int16_t Min(int16_t a, int16_t b) {
  return min(a,b);
}

//-----------------------------------------------------------
// Max()
//   Min funtion
//   it's a function not a macro like max() 
//   so args are not calculated twice
//-----------------------------------------------------------
int16_t Max(int16_t a, int16_t b) {
  return max(a,b);
}

//-----------------------------------------------------------
// Constrain()
//   Constrain funtion
//   it's a function not a macro like constrain() 
//   so args are not calculated twice
//-----------------------------------------------------------
int16_t Constrain(int16_t a, int16_t lo, int16_t hi) {
  return constrain(a,lo,hi);
}

//-----------------------------------------------------------
// ConstrainF()
//   ConstrainF funtion
//   it's a function not a macro like constrain() 
//   so args are not calculated twice
//-----------------------------------------------------------
float ConstrainF(float a, float lo, float hi) {
  return constrain(a,lo,hi);
}

//-----------------------------------------------------------
// MoveServos()
//   returns if less than 50mS have passed since last call
//   change servo position so they're closer to their target
//   sets AtTargets to say whether all servos are at their targets
//   AtTargets will also be false if less than 50mS have passed since last call
//   returns value of AtTargets 
//-----------------------------------------------------------
bool MoveServos() {
//  Serial.write(0xFF); // to flash the Tx LED  
  AtTargets = false;
  
  static unsigned long previousMillis = 0;
  if (millis() - previousMillis < VelPeriod)
    return false;
  previousMillis = millis();
  
  float g[NumServosMax];
  float a1,tmax,sign1,p0;
  const float Vmax0 = 1000;

  AtTargets = true;

  for (byte grp=0; grp < NumServos; grp++) {
    tmax = 0;
    for (byte servo=0; servo < NumServos; servo++)
    if (Servos[servo].Group == grp) {
      if (Servos[servo].Vmax == 0)
        g[servo] = abs(Servos[servo].Target-Servos[servo].Position)/Vmax0; else
        g[servo] = abs(Servos[servo].Target-Servos[servo].Position)/Servos[servo].Vmax;
      if (Servos[servo].Amax > 0) {
        sign1 = sign(sqr(Servos[servo].Velocity) + Servos[servo].Amax*2*(2*abs(Servos[servo].Velocity) + sign(Servos[servo].Velocity)*(Servos[servo].Position-Servos[servo].Target)));
        g[servo] = Max(g[servo],
          (sign1*abs(Servos[servo].Velocity) + sqrt(abs(2*sign1*Servos[servo].Velocity*abs(Servos[servo].Velocity) + 4*Servos[servo].Amax*(Servos[servo].Position-Servos[servo].Target))))/Servos[servo].Amax);
      }
      tmax = Max(tmax,g[servo]);
    }

    for (byte servo=0; servo < NumServos; servo++)
    if (Servos[servo].Group == grp) {
      if (grp == 0)
        tmax = g[servo];
        
      if ((tmax < 2) or ((abs(Servos[servo].Velocity) <= Servos[servo].Amax+0.0001) and (abs(Servos[servo].Position+Servos[servo].Velocity-Servos[servo].Target) <= 1))) {
        Servos[servo].Position = Servos[servo].Target;
        Servos[servo].Velocity = 0;
      } else
      if (tmax > 0) {
        g[servo] = g[servo]/tmax;
        a1 = Servos[servo].Amax*sqr(g[servo]);
  
        if (a1 == 0) {
          Servos[servo].Velocity = (Servos[servo].Target-Servos[servo].Position)/tmax;
        } else {
          p0 = 2*Servos[servo].Velocity + abs(Servos[servo].Velocity)*Servos[servo].Velocity/(2*a1);
          a1 = sign(Servos[servo].Target-Servos[servo].Position-p0)*a1;
          if (a1*(Servos[servo].Target-Servos[servo].Position) < 0) 
          {
            a1  =  sign(a1)*abs(Servos[servo].Velocity)*Servos[servo].Velocity/(2*(Servos[servo].Target-Servos[servo].Position));
            a1 = constrain(a1,-Servos[servo].Amax,Servos[servo].Amax);
          }
  
          Servos[servo].Velocity = Servos[servo].Velocity+a1;
          if (Servos[servo].Vmax > 0) {
            float v1 = Servos[servo].Vmax*g[servo];
            Servos[servo].Velocity = constrain(Servos[servo].Velocity,-v1,v1);
          }
        }

        if (Servos[servo].Vmax > 0) 
          Servos[servo].Velocity = constrain(Servos[servo].Velocity,-Servos[servo].Vmax,Servos[servo].Vmax);
        Servos[servo].Position = Servos[servo].Position+Servos[servo].Velocity;
      }
      
      AtTargets = AtTargets && (Servos[servo].Position == Servos[servo].Target);
    }
  }

  for (byte servo=0; servo < NumServos; servo++) {
    Servos[servo].Position = constrain(Servos[servo].Position,Servos[servo].ServoMin,Servos[servo].ServoMax);
    Servos[servo].aServo.writeMicroseconds(Servos[servo].Position);
    if (!Servos[servo].aServo.attached())
      Servos[servo].aServo.attach(Servos[servo].PinNum);
  }
  return AtTargets;
}

//-----------------------------------------------------------
// SetServoPosition
//   sets Target and Position immediately
//-----------------------------------------------------------
void SetServoPosition(uint8_t servo, int16_t pos) {
  Servos[servo].Target = pos;
  Servos[servo].Position = pos;
  Servos[servo].aServo.writeMicroseconds(pos);
  if (!Servos[servo].aServo.attached())
    Servos[servo].aServo.attach(Servos[servo].PinNum);
}

#endif
